test.event_mgmt                 = {}
test.event_mgmt.toggle          = {}
test.event_mgmt.build_dialog    = {}
test.event_mgmt.ui              = {}
test.event_mgmt.ui.dialog       = {}

local em = test.event_mgmt
local ui = em.ui

em.minor = 'eventmanager'

--                                test.events[num index] 
--                                              ^   ^
--                                              |   : 
--  test.events_ref.reverse_map[event_name key]-/   ...test.events_ref.registered[num index]
--                                     |
--                                     v
--     test.events_ref.iup_controls[num index]

-- test.events_ref.reverse_map holds a numeric index corresponding to the event_name, used as a key, for accessing
--     the event within test.events and the iup.toggle within test.events_ref.iup_controls

-- test.events_ref.registered is accessed via the numeric index of the event and holds a reference to the data
--     in test.events

local state_map = {ON = true, OFF = false}
local bool_map  = {[true] = 'ON', [false] = 'OFF'}

function test.event_mgmt.init()
	em.load_defaults(true)
	em.build_dialog()
end

function test.event_mgmt.load_defaults(init)
	local notes = unspickle(gkini.ReadString(test.config.major, em.minor..'.defaults', ''))
--	if (#notes ~= 0) then test.events = notes else test.events = test.events_default end
	-- Make sure that no events are currently registered
	if (assert(function() for k in pairs(test.events_ref.registered) do return true end return false end)()) then
		em.unregister_all()
	end

	-- Register defaults
		for _, v in ipairs(notes) do
			em.manual_toggle('reg', test.events[v][1], true, true)
		end

		--if (state_map[v[2]]) then em.register_events(v) end

end

function test.event_mgmt.readevent(event, ...)
	if (not test.debug.events) then return end
	local data = {...}
	local str = ''

	for n = 1, select('#',...) do
		local e = select(n,...)
		local t = type(e)
		str = str..'\n\t'..n..': ('..t..')\t'..test.stringify(e)
	end

-- 	for k,v in pairs(data) do
-- 		if (type(v) == "nil") then v = "nil"
-- 		elseif (type(v) == "table") then v = spickle(v)
-- 		end
-- 		str = str..'\n\t'..tostring(v)
-- 	end
	console_print("Event: "..event.." - Data:"..str)
end

function test.event_mgmt.register_events(tbl)
	-- Accepts a table and (un)registers an event based on the value of the second index
	local event, state = unpack(tbl)
	local idx = test.events_ref.reverse_map[event]

	if (state_map[state]) then
		RegisterEvent(test.event_mgmt.readevent, event)
		test.events_ref.registered[idx] = test.events[idx]
	else
		UnregisterEvent(test.event_mgmt.readevent, event)
		test.events_ref.registered[idx] = nil
	end
end

function test.event_mgmt.manual_toggle(action, event, silent, init)
-- manual_toggle(action, event [,silent [, init]])
	--event = event:upper()
	--if (event == 'RHUDXSCALE') then event = 'rHUDxscale' end
	local idx = test.events_ref.reverse_map[event]

	if (idx == nil) then print('Unknown event name. ('..event..')'); return end

	if (action == 'reg') then
		if (not state_map[test.events[idx][2]] or init) then 
			test.events[idx][2] = 'ON'
			pcall(function() test.events_ref.iup_controls[idx].value = test.events[idx][2] end)
			em.register_events(test.events[idx])
		else
			action = 'is already '..action
		end
	elseif (action == 'unreg') then
		if (state_map[test.events[idx][2]]) then 
			test.events[idx][2] = 'OFF'
			test.events_ref.iup_controls[idx].value = test.events[idx][2]
			em.register_events(test.events[idx])
		else
			action = 'is already '..action
		end
	end
	if (not silent) then print('Event '..action..'istered: '..event) end
end

function test.event_mgmt.event_status(getlistonly)
--event_status([getlistonly])
	local count = 0
	local event_list = {}

	for _,v in pairs(test.events_ref.registered) do
		table.insert(event_list, v[1])
		count = count + 1
	end

	if (not getlistonly) then
		print('The list of registered events has been printed to the console.')
		console_print('Registered Events\n'..
					'---------------------------')
		if (count == 0) then 
			console_print('\tnone')
		else
			console_print('\t'..table.concat(event_list, '\n\t')) 
		end
	end
	return event_list
end

function test.event_mgmt.unregister_all()
	local reg_events = em.event_status(true)
	for _, v in ipairs(reg_events) do
		em.manual_toggle('unreg', v, true)
	end
end
-- ----------------------------------------------------------------------------

function test.event_mgmt.toggle:New(tbl)
	local event, state = unpack(tbl)
	local tog = iup.stationtoggle{title = event, value = state}
	
	function tog:action()
		local idx = test.events_ref.reverse_map[self.title]
		test.events[idx][2] = self.value
		em.register_events(test.events[idx])
	end
	return tog
end

test.event_mgmt.ui.debugevents = iup.stationtoggle{title = 'Debug Events', action = function(self, v)
		if (v == nil) then 
			self.value = bool_map[test.debug.events]
		else
			test.debug.events = state_map[self.value]
		end
		gkini.WriteInt(test.config.major, 'debug.events', test.debug.events and 1 or 0)
	end
}

test.event_mgmt.ui.selectall = iup.stationbutton{title = 'Select All', action = function()
		for _, v in ipairs(test.events) do
			em.manual_toggle('reg', v[1], true)
		end
	end
}

test.event_mgmt.ui.clearall = iup.stationbutton{title = 'Clear All', action = function()
		em.unregister_all()
	end
}

test.event_mgmt.ui.load = iup.stationbutton{title = 'Load Defaults', action = function()
		em.unregister_all()
		em.load_defaults()
	end
}

test.event_mgmt.ui.save = iup.stationbutton{title = 'Save Defaults', action = function()
		local notes = {}
		for _, v in ipairs(em.event_status(true)) do
			table.insert(notes, test.events_ref.reverse_map[v])
		end
		gkini.WriteString(test.config.major, em.minor..'.defaults', spickle(notes))
	end
}

test.event_mgmt.ui.close = iup.stationbutton{title="Close", expand="HORIZONTAL", action = function(self)
		HideDialog(ui.dialog)
	end
}

test.event_mgmt.ui.listbox = iup.stationsublist{{}, control = "YES", expand = "YES",}

function test.event_mgmt.build_dialog()
	local listbox = ui.listbox
	local last_char, cur_char
	local hbox = {}
	local list = iup.vbox{}
	local vbox

	local function subdlg(ctrl)
		local dlg = iup.dialog{
			ctrl,
			border="NO",
			menubox="NO",
			resize="NO",
			bgcolor="0 0 0 0 *",
		}
		return dlg
	end

	-- Create list of iup.toggle elements for each event
	for i, v in ipairs(test.events) do
		-- Get and check first letter for section label
		cur_char = string.upper(string.match(v[1], "^[%a%u]"))
		if (cur_char ~= last_char) then 
			if (last_char) then -- Add end-of-section spacing
				iup.Append(list, vbox)
				iup.Append(list, iup.fill{size = 25}) 
			end
			--Add new section label
			last_char = cur_char
			vbox = iup.vbox{gap = 2}
			iup.Append(vbox, iup.hbox{iup.label{title = cur_char}, iup.fill{}})
		end
		test.events_ref.iup_controls[i] = em.toggle:New(v)
		iup.Append(vbox, iup.hbox{test.events_ref.iup_controls[i], iup.fill{}}) 
	end
	iup.Append(list, vbox)

	iup.Append(listbox, subdlg(list))
	listbox[1] = 1

	ui.debugevents.value = bool_map[test.debug.events]

	ui.main = iup.pdarootframe{
		iup.pdarootframebg{
			iup.vbox{
				iup.hbox{
					iup.label{title="Events"},
					iup.fill{expand = 'YES'},
					ui.debugevents,
					iup.fill{size = 20},
					ui.selectall,
					ui.clearall,
					iup.fill{size = 20},
					ui.load,
					ui.save,
					gap = 10,
				},
				listbox,
				ui.close,
				gap=5,
			},
		},
		size="TWOTHIRDxTWOTHIRD", expand="NO",
	}

	ui.dialog = iup.dialog{
		iup.vbox{
			iup.fill{},
			iup.hbox{
				iup.fill{},
				ui.main,
                iup.fill{},
			},
			iup.fill{},
		},
		defaultesc=ui.close,
		bgcolor="0 0 0 128 *",
		fullscreen="YES",
		border="NO",
		resize="NO",
		menubox="NO",
		topmost="YES",
	}
	ui.dialog:map()
end

function test.event_mgmt.ui.Open()
	ShowDialog(ui.dialog)
end
-- function test.event_mgmt.ui.Close()
-- 	HideDialog(test.event_mgmt.ui.dialog)
-- end
-- test.event_mgmt.ui.close.action = test.event_mgmt.ui.Close
