test.xor                = {}
test.nz                 = {}
test.exists             = {}
test.IsIupControl       = {}
test.IsIupContainer     = {}
test.strsplit           = {}
test.stringify          = function() end

local tinsert = table.insert
local tremove = table.remove

function test.xor(a, b) return ((a and not b) or (not a and b)) end

function test.nz(val, onnil)
	if (val == nil) then return (onnil or 'nil') else return val end
end

function test.print(str)
	if (test.debug.print) then console_print(str) end
end

local console_print_buffer = {}
local console_print_timer = Timer()
function test.buffered_console_print(str, delay)
	delay = delay or test.buffered_console_print_delay

	local function print_buffer()
		local buffer = console_print_buffer
		if (#buffer > 0) then console_print(buffer[1]) else return end
		if (#buffer > 0) then 
			tremove(buffer, 1)
			if (test.buffered_console_print_delay > 0) then test.print('There are '..tostring(#buffer)..' entries left.') end
			console_print_timer:SetTimeout(delay, print_buffer)
		end
	end
	tinsert(console_print_buffer, str)
	if (not console_print_timer:IsActive()) then print_buffer() end
end
	
function test.exists(val)
	if not val then return nil end
	return pcall(loadstring('return '..val))
	-- if (val) then return true, loadstring("return "..val)() end
end

function test.IsIupControl(control)
	return pcall(iup.GetType,control)
end

function test.IsIupContainer(control)
	if (type(control) ~= 'userdata') then return false end
	return test.container_types[iup.GetType(control)]
end


function test.stringify(v, nobinary, count)
--Returns the string value of the parameter passed to it.
--Works with nspickle() to display the contents of tables that contain binary data
--Setting nobinary to true will strip binary data from the string form of tables.
	local function nspickle(tbl, nobinary, count)
		local msg = ''
		local n = 0

		for i, v in ipairs(tbl) do
			local tv = type(v)
			if (not (nobinary and (tv == 'function' or tv == 'userdata' or tv == 'thread'))) then 
				msg = msg..test.stringify(v, nobinary, count)..','
				n = i
			else
				break
			end
		end
		msg = msg:gsub(',$', ';')
		for k, v in pairs(tbl) do
			local tk = type(k)
			local tv = type(v)
			if (tk ~= 'number' or (tk == 'number' and (k < 1 or k > n))) then
				if (tk == 'number') then 
					k = '['..k..']'
				elseif (tk == 'string' and k:match('[^%w_]')) then 
					k = '["'..k..'"]'
				end
				if (not (nobinary and (tv == 'function' or tv == 'userdata' or tv == 'thread'))) then 
					msg = msg..tostring(k)..'='..test.stringify(v, nobinary, count)..',' 
				end
			end
		end
		msg = msg:gsub('[,;]$', '')
		return msg
	end

	local t = type(v)
	if (t == 'table') then 
		count = (count or 0) + 1
		v = '{'..nspickle(v, nobinary, count)..'}'
		count = count - 1
		if (count == 0) then v = v:gsub('^{(.*)}$', '%1') end
	elseif (t == 'string') then v = '"'..v..'"'
	else v = tostring(v) 
	end
	return v
end

function test.strsplit(pattern,text)
    local result = {}
    local theStart = 1
    local theSplitStart, theSplitEnd = string.find(text,pattern,theStart)
    while theSplitStart do
        table.insert(result,string.sub(text,theStart,theSplitStart-1))
        theStart = theSplitEnd + 1
        theSplitStart, theSplitEnd = string.find(text,pattern,theStart)
    end
    table.insert(result, string.sub(text,theStart))
    return result 
end

function test.jointable(tbl, i, j)
-- joins a table containing component parts of a table's name 
	if (type(tbl) ~= 'table') then return nil, 'Expected a table value' end
	i = i or 1
	j = j or #tbl
	local str = table.concat(tbl, '.', i, j)
	return str:gsub('%.%[', '[')
end

function test.splittable(str)
-- splits a given string representing a table name into it's component parts
	if (type(str) ~= 'string') then return nil, 'Expected a string value' end
	local parts = test.strsplit('%.', str)
	local ni = {}
	for i, v in ipairs(parts) do
		parts[i] = v:gsub('(%[.-%])', function(c)
			if (ni[i] == nil) then ni[i] = {} end
			ni.n = i
			table.insert(ni[i], c)
			return ""
		end)
	end
	do
		local inc = 0
		for i=1, (ni.n or 0) do
			if (ni[i] ~= nil) then
				for j, w in ipairs(ni[i]) do
					table.insert(parts, i+j+inc, w)
				end
			end
			inc = inc + #(ni[i] or {})
		end
	end
	return parts
end

function test.list2(params)
	-- call this function with table notation.  list2{}
	params.expand = params.expand or 'NO'
	params.list = params.list or {}
	params.size = params.size or 50
	params.value = params.value or ''
	local textbox = iup.text{size = params.size, expand = params.expand, value = params.value}
	local dropdown = iup.list{dropdown = 'YES', size = params.size + 20, expand = params.expand, fgcolor = '0 0 0 0', action = function(self, t, i, v)
		textbox.value = t
		iup.SetFocus(textbox)
	end}
	for i, v in ipairs(params.list) do dropdown[i] = v end
	local zbox = iup.zbox{
		--dropdown must be before textbox within the zbox to create the proper z-order
		iup.hbox{dropdown,},
		iup.hbox{textbox, iup.fill{size = 20}},
		all = 'YES',
		alignment = 'NW',
	}
	return zbox, textbox, dropdown
end

test.newfunc = function(a, b) print('a='..tostring(a)..' / b='..tostring(b)) return 5, 6 end
